/**
  * Copyright (c): Uwe Schmidt, FH Wedel
  *
  * You may study, modify and distribute this source code
  * FOR NON-COMMERCIAL PURPOSES ONLY.
  * This copyright message has to remain unchanged.
  *
  * Note that this document is provided 'as is',
  * WITHOUT WARRANTY of any kind either expressed or implied.
  */
/**
  * The original sourcecode was modified to 
  * fit the phonebook requirements.
  * This sourcefile only contains the required 
  * functionality for the xml-example. For 
  * binary-tree-implementation, please check the
  * original sourcecode.
  * This file may contain some additional functions.
  * 
  * @author: Sascha Wille (inf6201@fh-wedel.de)
  */

#include "phonebook.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

/*
 * create an empty phonebook
 *
 * @return empty phonebook
 */
Phonebook mkEmpty(void) {
    return (Phonebook) 0;
}

/*
 * Determines wheather phonebook is empty or not
 *
 * @param[in] phonebook to check
 *
 * @return boolean 1 if phonebook is empty 0 otherwise
 */
int isEmpty(Phonebook pb) {
    return pb == (Phonebook) 0;
}

/*
 * create a phonebook with one entry
 *
 * @param[in] entry of phonebook
 *
 * @return phonebook with exactly one entry
 */
Phonebook mkOne(Entry e) {
    Phonebook res = malloc(sizeof (*res));

    if (!res)
        exit(1);
    res->info = e;
    res->l = mkEmpty();
    res->r = mkEmpty();

    return res;
}

/*
 * Determines wheather entry is in phonebook or not
 *
 * @param[in] entry to check if it's in
 * @param[in] base
 *
 * @return boolean 1 if Entry was in 0 otherwise
 */
int isIn(Entry e, Phonebook pb) {
    if (isEmpty(pb))
        return 0;

    switch (compare(e, pb->info)) {
        case -1 :
            return isIn(e, pb->l);
        case 0:
            return 1;
        case +1 :
            return isIn(e, pb->r);
    }
    return 0;
}

/*
 * adds an entry into the phonebook
 *
 * @param[in] additional entry
 *
 * @return changed phonebook
 */
Phonebook addEntry(Entry e, Phonebook pb) {
    if (isEmpty(pb))
        return mkOne(e);

    switch (compare(e, pb->info)) {
        case -1 :
            pb->l = addEntry(e, pb->l);
            break;
        case 0:
            break;
        case +1 :
            pb->r = addEntry(e, pb->r);
            break;
    }

    return pb;
}

/*
 * Ammount of entries in a phonebook
 *
 * @param[in] phonebook to check
 *
 * @return ammount of entries
 */
unsigned int size(Phonebook pb) {
    if (isEmpty(pb)) return 0;
    return 1 + size(pb->l) + size(pb->r);
}

/*
 * Returns the greatest value in the tree
 *
 * parm[in] phonebook to compute
 *
 * @return greatest entry in phonebook
 */
static Entry maxEntry(Phonebook pb) {
    assert(!isEmpty(pb));

    if (isEmpty(pb->r))
        return pb->info;

    return maxEntry(pb->r);
}

/*
 * Removes the root entry of the phonebook
 *
 * parm[in] phonebook to compute
 *
 * @return phonebook with "new" root element
 */
static Phonebook removeRoot(Phonebook pb) {
    assert(!isEmpty(pb));

    if (isEmpty(pb->l)
            || isEmpty(pb->r)) {
        Phonebook res =
                isEmpty(pb->l) ? pb->r : pb->l;
        free(pb);
        return res;
    }

    assert(!isEmpty(pb->l));
    assert(!isEmpty(pb->r));

    pb->info = maxEntry(pb->l);
    pb->l = removeEntry(pb->info, pb->l);

    return pb;
}

/*
 * removes an entry into the phonebook
 *
 * @param[in] entry to be removed
 *
 * @return changed phonebook
 */
Phonebook removeEntry(Entry e, Phonebook pb) {
    if (isEmpty(pb))
        return pb;

    switch (compare(e, pb->info)) {
        case -1 :
            pb->l = removeEntry(e, pb->l);
            break;
        case 0:
            pb = removeRoot(pb);
            break;
        case +1 :
            pb->r = removeEntry(e, pb->r);
            break;
    }

    return pb;
}
/**
 * returns the number of entries in the phonebook
 *
 * @param[in] pb, phonebook
 * @return no of entries
 */
int getSize(Phonebook pb) {
    if (isEmpty(pb))
        return 0;
    return getSize(pb->l) + 1 + getSize(pb->r);
}

/**
 * builds spaces
 *
 * @param[in] n, number of spaces
 * @param[in] stream, outputstream
 */
static void
spaces(int n, FILE * stream){
    int i = 0;
    for (i = 0; i<n; ++i)
        fprintf(stream, " ");
}

/**
 * maximal value
 *
 * @param[in] v1, value1
 * @param[in] v2, value2
 *
 * @return max value
 */
static int max(int v1, int v2) {
    return v1 > v2 ? v1 : v2;
}

/**
 * prints the whole phonebook in 
 * lexicographical order
 *
 * @param[in] pb, phonebook
 */
void printPhoneBook(Phonebook pb) {
    if (isEmpty(pb))
        return;
    printPhoneBook(pb->l);

    fprintf(stdout, "Name: %s", pb->info->name);
    spaces(max(25 - strlen(pb->info->name), 0),stdout);

    fprintf(stdout, "\tFirstname: %s", (pb->info->firstname == NULL) ? "-" : pb->info->firstname);
    spaces(max(15 - ((pb->info->firstname == NULL) ? 1 : strlen(pb->info->firstname)), 0), stdout);

    fprintf(stdout, "\tPhone: %s", pb->info->phone);
    spaces(max(20 - strlen(pb->info->phone), 0), stdout);

    fprintf(stdout, "\tMobile: %s\n", (pb->info->mobile) ? "yes" : "no");
    printPhoneBook(pb->r);
}

/**
 * retieves an entry due to a phonenumber
 * ASSUMPTION! Phonenumbers are unique
 *
 * @param[in] phonenumber
 * @param[in] phonebook
 *
 * @return entry if it was in; NULL otherwise
 */
Entry getEntryByPhone(char * no, Phonebook pb) {
    if (isEmpty(pb))
        return NULL;

    if (!comparePhoneNo(no, pb->info))
        return pb->info;
    else {
        Entry e = getEntryByPhone(no, pb->l);
        return e ? e : getEntryByPhone(no, pb->r);
    }
}
