/**
 * Embedding Python Example (XML-Processing).
 * Copyright (c): Sascha Wille, inf6201@fh-wedel.de
 *
 * You may study, modify and distribute this source code
 * FOR NON-COMMERCIAL PURPOSES ONLY.
 * This copyright message has to remain unchanged.
 *
 * Note that this document is provided 'as is',
 * WITHOUT WARRANTY of any kind either expressed or implied.
 */

#include <Python.h>
#include "phonebook.h"

/**
 * This function parses an argument-array and extracts the phonenumber for
 * phonebook-lookup if applicable
 *
 * @param[in] argc, commandline arguments
 * @param[out] searchString, extracted phonenumber or NULL
 *
 * @return boolean whether searchString contains result or not
 */
unsigned int getSearch(char *args[], char **searchString) {
    char dummy[2] = {'\0', '\0'};
    char option = '\0';
    unsigned int res = 0;
    assert(args[2] && args[3]);
    if ((sscanf(args[2], "-%c%c", &option, &(dummy[0])) == 1)
            && (option == 's')
            && (sscanf(args[2], "%c%c", &(dummy[0]), &(dummy[1])) > 0)) {
        *searchString = args[3];
        res = 1;
    } else {
        *searchString = NULL;
    }
    return res;
}

/**
 * Add the local path to to python-system path
 *
 * @param[in] current path
 */
void setPath(wchar_t* path) {
    wchar_t newPath[1024] = L":.\0";
    wcscat(newPath, path);
    PySys_SetPath(newPath);
}

/**
 * Malloc's enough memory and performs a string copy
 *
 * @param[out] dest, source-string
 * @param[in] src, source-string
 */
void clone(char **dst, const char *src) {
    *dst = malloc(sizeof (char) * strlen(src) + 1);
    if (!dst) {
        exit(1);
    }
    strcpy(*dst, src);
}

/**
 * Loads a phonebook from a xml-file, specified by filename
 * 
 * @param[in] specifies the file to import
 *
 * @return phonebook (represented by tree)
 */
Phonebook loadPhoneBookFromXML(char *filename) {
    /* phonebook */
    Phonebook res = NULL;
    char *cName = NULL, *cFirstname = NULL, *cPhone = NULL, *ccName = NULL, *ccFirstname = NULL, *ccPhone = NULL;
    unsigned int cMobile;

    PyObject *phone = NULL, *values = NULL, *module = NULL, *className = NULL, *dict = NULL, *instance = NULL;

    /* Iterator-Startposition */
    Py_ssize_t pos;

    /* Modulname */
    PyObject *mName = PyUnicode_FromString("xmlwork");

    /* Python-Interpreter initialisieren */
    Py_Initialize();

    /* aktuelles Arbeitsverzeichnis in Interpreter-Suchpfad eintragen */
    setPath(Py_GetPath());

    /* Modul importieren */
    module = PyImport_Import(mName);

    if (module) {
        className = PyObject_GetAttrString(module, "Xml"); /* Objekt mit Namen Xml holen */

        if (className && PyCallable_Check(className)) { /* lässt sich die Klasse instanzieren? */

            instance = PyObject_CallFunction(className, "(s)", filename); /* Konstruktoraufruf */
            if (instance) {
                dict = PyObject_CallMethod(instance, "load_address_book", "");

                if (dict) {
                    fprintf(stdout, "Python-Import finished\n%d Entries imported\n\n", (int) PyDict_Size(dict));

                    /* Leeren Baum erstellen */
                    res = mkEmpty();

                    /* Iteration ueber Dictionary (nur geborgte Referenzen) */
                    pos = 0;
                    while (PyDict_Next(dict, &pos, &phone, &values)) {
                        /* Tupel auseinandernehmen (Name, Vorname/None, Mobil(Ja/Nein) */
                        PyArg_ParseTuple(values, "szi", &cName, &cFirstname, &cMobile); /* Formatstring z: String oder None */

                        /* Vornamen aus Dictionary holen */
                        if (cFirstname) {
                            clone(&ccFirstname, cFirstname);
                        }

                        /* Telefonnummer aus Schluesselfeld extrahieren */
                        PyArg_Parse(phone, "s", &cPhone);

                        /* Strings fuer spaetere Verwendung kopieren */
                        clone(&ccPhone, cPhone);
                        clone(&ccName, cName);

                        /* Alle Werte in den Baum einfuegen */
                        res = addEntry(mkEntry(ccName, ccFirstname, ccPhone, cMobile), res);
                        ccFirstname = NULL; /* firstname is allowed to be NULL */
                    }
                } else {
                    fprintf(stderr, "Adressbuch %s konnte nicht verarbeitet werden.\n", filename);
                }
            }
        } else {
            fprintf(stderr, "Keine Instanziierung vom Typ Xml möglich\n");
            PyErr_Clear(); /* do not pass error to the interpreter */
        }
    } else {
        fprintf(stderr, "Fehler beim Laden des Moduls xmlwork.py\n");
    }

    /* Den Refcount aller "eigenen" Referenzen dekrementieren */
    Py_XDECREF(className);
    Py_XDECREF(instance);
    Py_XDECREF(dict);
    Py_XDECREF(module);
    Py_XDECREF(mName);

    /* Shutdown interpreter (free, ...) */
    Py_Finalize();

    return res;
}

int main(int argc, char ** argv) {
    unsigned int res = 0;
    char * search = NULL;
    Entry e = NULL;

    /* Phonebook */
    Phonebook pb = loadPhoneBookFromXML(argv[1]);

    if (pb) {
        printPhoneBook(pb);
        fprintf(stdout,"\nPhonebook contains %d elements\n", getSize(pb));

        if (argc == 4) {
            res = getSearch(argv, &search);
            switch (res) {
                    /* Backward search */
                case 1: e = getEntryByPhone(search, pb);
                    break;
                    /* Forwardsearch not implemeted yet */
                case 2:
                    break;
            }
        }
        if (e)
            printEntry(e);
        else if (res)
            fprintf(stdout, "\n\nEintrag im Adressbuch nicht gefunden.\n");

    } else {
        fprintf(stderr, "Adressbuch konnte nicht geladen werden.\n");
    }

    return 0;
}
